@@ -21,8 +21,10 @@
 #include "rpl_constants.h"
 
 class Relay_log_info;
+class Gtid_index_writer;
 
 class Format_description_log_event;
+class Gtid_log_event;
 
 bool reopen_fstreams(const char *filename, FILE *outstream, FILE *errstream);
 void setup_log_handling();
@@ -240,6 +242,7 @@ extern TC_LOG_DUMMY tc_log_dummy;
 #define LOG_CLOSE_TO_BE_OPENED	2
 #define LOG_CLOSE_STOP_EVENT	4
 #define LOG_CLOSE_DELAYED_CLOSE 8
+#define LOG_CLOSE_SYNC_GTID_INDEX 16
 
 /* 
   Maximum unique log filename extension.
@@ -257,7 +260,7 @@ extern TC_LOG_DUMMY tc_log_dummy;
 class Relay_log_info;
 
 /*
-  Note that we destroy the lock mutex in the desctructor here.
+  Note that we destroy the lock mutex in the destructor here.
   This means that object instances cannot be destroyed/go out of scope,
   until we have reset thd->current_linfo to NULL;
  */
@@ -403,7 +406,7 @@ public:
   void set_write_error(THD *thd, bool is_transactional);
   static bool check_write_error(THD *thd);
   static bool check_cache_error(THD *thd, binlog_cache_data *cache_data);
-  int write_cache(THD *thd, IO_CACHE *cache);
+  int write_cache(THD *thd, binlog_cache_data *cache_data);
   int write_cache_raw(THD *thd, IO_CACHE *cache);
   char* get_name() { return name; }
   void cleanup()
@@ -429,6 +432,8 @@ public:
                                    bool is_relay_log);
 
   bool write_event(Log_event *ev, binlog_cache_data *data, IO_CACHE *file);
+  bool write_event(Log_event *ev, enum enum_binlog_checksum_alg checksum_alg,
+                   binlog_cache_data *data, IO_CACHE *file);
 };
 
 /**
@@ -595,9 +600,12 @@ class binlog_cache_mngr;
 class binlog_cache_data;
 struct rpl_gtid;
 struct wait_for_commit;
+class Binlog_commit_by_rotate;
 
 class MYSQL_BIN_LOG: public TC_LOG, private Event_log
 {
+  friend Binlog_commit_by_rotate;
+
 #ifdef HAVE_PSI_INTERFACE
   /** The instrumentation key to use for @ LOCK_index. */
   PSI_mutex_key m_key_LOCK_index;
@@ -675,6 +683,7 @@ class MYSQL_BIN_LOG: public TC_LOG, private Event_log
   mysql_cond_t  COND_xid_list;
   mysql_cond_t  COND_relay_log_updated, COND_bin_log_updated;
   ulonglong bytes_written;
+  ulonglong binlog_space_total;
   IO_CACHE index_file;
   char index_file_name[FN_REFLEN];
   /*
@@ -720,6 +729,9 @@ class MYSQL_BIN_LOG: public TC_LOG, private Event_log
   ulonglong group_commit_trigger_count, group_commit_trigger_timeout;
   ulonglong group_commit_trigger_lock_wait;
 
+  /* Binlog GTID index. */
+  Gtid_index_writer *gtid_index;
+
   /* pointer to the sync period variable, for binlog this will be
      sync_binlog_period, for relay log this will be
      sync_relay_log_period
@@ -729,6 +741,13 @@ class MYSQL_BIN_LOG: public TC_LOG, private Event_log
   bool state_file_deleted;
   bool binlog_state_recover_done;
 
+  Gtid_index_writer *recover_gtid_index_start(const char *base_name,
+                                              my_off_t offset);
+  void recover_gtid_index_process(Gtid_index_writer *gi, my_off_t offset,
+                                  const rpl_gtid *gtid);
+  void recover_gtid_index_end(Gtid_index_writer *gi);
+  void recover_gtid_index_abort(Gtid_index_writer *gi);
+
   inline uint get_sync_period()
   {
     return *sync_period_ptr;
@@ -740,16 +759,24 @@ class MYSQL_BIN_LOG: public TC_LOG, private Event_log
     new_file() is locking. new_file_without_locking() does not acquire
     LOCK_log.
   */
-  int new_file_impl();
+  int new_file_impl(bool commit_by_rotate);
   void do_checkpoint_request(ulong binlog_id);
-  void purge();
-  int write_transaction_or_stmt(group_commit_entry *entry, uint64 commit_id);
+  int write_transaction_or_stmt(group_commit_entry *entry, uint64 commit_id,
+                                bool commit_by_rotate);
   int queue_for_group_commit(group_commit_entry *entry);
   bool write_transaction_to_binlog_events(group_commit_entry *entry);
+  bool write_transaction_with_group_commit(group_commit_entry *entry);
+  void write_transaction_handle_error(group_commit_entry *entry);
   void trx_group_commit_leader(group_commit_entry *leader);
+  void trx_group_commit_with_engines(group_commit_entry *leader,
+                                     group_commit_entry *tail,
+                                     bool commit_by_rotate);
   bool is_xidlist_idle_nolock();
+  void update_gtid_index(uint32 offset, rpl_gtid gtid);
+
 public:
-  int new_file_without_locking();
+  void purge(bool all);
+  int new_file_without_locking(bool commit_by_rotate);
   /*
     A list of struct xid_count_per_binlog is used to keep track of how many
     XIDs are in prepared, but not committed, state in each binlog. And how
@@ -768,11 +795,8 @@ public:
     ulong binlog_id;
     /* Total prepared XIDs and pending checkpoint requests in this binlog. */
     long xid_count;
-    long notify_count;
-    /* For linking in requests to the binlog background thread. */
-    xid_count_per_binlog *next_in_queue;
     xid_count_per_binlog(char *log_file_name, uint log_file_name_len)
-      :binlog_id(0), xid_count(0), notify_count(0)
+      :binlog_id(0), xid_count(0)
     {
       binlog_name_len= log_file_name_len;
       binlog_name= (char *) my_malloc(PSI_INSTRUMENT_ME, binlog_name_len, MYF(MY_ZEROFILL));
@@ -814,7 +838,7 @@ public:
     FD     - Format-Description event,
     R      - Rotate event
     R_f    - the fake Rotate event
-    E      - an arbirary event
+    E      - an arbitrary event
 
     The underscore indexes for any event
     `_s'   indicates the event is generated by Slave
@@ -982,7 +1006,8 @@ public:
 	    enum cache_type io_cache_type_arg,
 	    ulong max_size,
             bool null_created,
-            bool need_mutex);
+            bool need_mutex,
+            bool commit_by_rotate = false);
   bool open_index_file(const char *index_file_name_arg,
                        const char *log_name, bool need_mutex);
   /* Use this to start writing a new log file */
@@ -1006,19 +1031,24 @@ public:
 
   using Event_log::write_event;
 
-  bool write_event(Log_event *ev) { return write_event(ev, 0, &log_file); }
+  bool write_event(Log_event *ev, enum enum_binlog_checksum_alg checksum_alg)
+  {
+    return write_event(ev, checksum_alg, 0, &log_file);
+  }
+  bool write_event(Log_event *ev);
 
   bool write_event_buffer(uchar* buf,uint len);
-  bool append(Log_event* ev);
-  bool append_no_lock(Log_event* ev);
+  bool append(Log_event* ev, enum enum_binlog_checksum_alg checksum_alg);
+  bool append_no_lock(Log_event* ev, enum enum_binlog_checksum_alg checksum_alg);
 
   void mark_xids_active(ulong cookie, uint xid_count);
   void mark_xid_done(ulong cookie, bool write_checkpoint);
   void make_log_name(char* buf, const char* log_ident);
   bool is_active(const char* log_file_name);
-  bool can_purge_log(const char *log_file_name);
+  bool can_purge_log(const char *log_file_name, bool interactive);
   int update_log_index(LOG_INFO* linfo, bool need_update_threads);
-  int rotate(bool force_rotate, bool* check_purge);
+  int rotate(bool force_rotate, bool *check_purge,
+             bool commit_by_rotate= false);
   void checkpoint_and_purge(ulong binlog_id);
   int rotate_and_purge(bool force_rotate, DYNAMIC_ARRAY* drop_gtid_domain= NULL);
   /**
@@ -1035,11 +1065,27 @@ public:
      @retval other Failure
   */
   bool flush_and_sync(bool *synced);
-  int purge_logs(const char *to_log, bool included,
-                 bool need_mutex, bool need_update_threads,
+  int purge_logs(THD *thd, const char *to_log, bool included,
+                 bool need_mutex, bool need_update_threads, bool interactive,
                  ulonglong *decrease_log_space);
-  int purge_logs_before_date(time_t purge_time);
+  int purge_logs_before_date(THD *thd, time_t purge_time, bool interactive);
   int purge_first_log(Relay_log_info* rli, bool included);
+  int count_binlog_space();
+  void count_binlog_space_with_mutex()
+  {
+    mysql_mutex_lock(&LOCK_index);
+    count_binlog_space();
+    mysql_mutex_unlock(&LOCK_index);
+  }
+  ulonglong get_binlog_space_total();
+  int real_purge_logs_by_size(ulonglong binlog_pos);
+  inline int purge_logs_by_size(ulonglong binlog_pos)
+  {
+    if (is_relay_log || ! binlog_space_limit ||
+        binlog_space_total + binlog_pos <= binlog_space_limit)
+      return 0;
+    return real_purge_logs_by_size(binlog_pos);
+  }
   int set_purge_index_file_name(const char *base_file_name);
   int open_purge_index_file(bool destroy);
   bool truncate_and_remove_binlogs(const char *truncate_file,
@@ -1082,6 +1128,7 @@ public:
   bool is_xidlist_idle();
   bool write_gtid_event(THD *thd, bool standalone, bool is_transactional,
                         uint64 commit_id,
+                        bool commit_by_rotate,
                         bool has_xid= false, bool ro_1pc= false);
   int read_state_from_file();
   int write_state_to_file();
@@ -1166,6 +1213,7 @@ public:
                         const char *user_host, size_t user_host_len, ulonglong query_utime,
                         ulonglong lock_utime, bool is_command,
                         const char *sql_text, size_t sql_text_len)= 0;
+  ATTRIBUTE_FORMAT(printf, 3, 0)
   virtual bool log_error(enum loglevel level, const char *format,
                          va_list args)= 0;
   virtual bool log_general(THD *thd, my_hrtime_t event_time, const char *user_host, size_t user_host_len, my_thread_id thread_id,
@@ -1193,6 +1241,7 @@ public:
                         const char *user_host, size_t user_host_len, ulonglong query_utime,
                         ulonglong lock_utime, bool is_command,
                         const char *sql_text, size_t sql_text_len) override;
+  ATTRIBUTE_FORMAT(printf, 3, 0)
   bool log_error(enum loglevel level, const char *format,
                          va_list args) override;
   bool log_general(THD *thd, my_hrtime_t event_time, const char *user_host, size_t user_host_len, my_thread_id thread_id,
@@ -1223,6 +1272,7 @@ public:
                         const char *user_host, size_t user_host_len, ulonglong query_utime,
                         ulonglong lock_utime, bool is_command,
                         const char *sql_text, size_t sql_text_len) override;
+  ATTRIBUTE_FORMAT(printf, 3, 0)
   bool log_error(enum loglevel level, const char *format,
                          va_list args) override;
   bool log_general(THD *thd, my_hrtime_t event_time, const char *user_host, size_t user_host_len, my_thread_id thread_id,
@@ -1279,10 +1329,12 @@ public:
   void cleanup_base();
   /* Free memory. Nothing could be logged after this function is called */
   void cleanup_end();
+  ATTRIBUTE_FORMAT(printf, 3, 0) // 1st arg is the implicit `this`
   bool error_log_print(enum loglevel level, const char *format,
                       va_list args);
   bool slow_log_print(THD *thd, const char *query, size_t query_length,
                       ulonglong current_utime);
+  ATTRIBUTE_FORMAT(printf, 4, 0)
   bool general_log_print(THD *thd,enum enum_server_command command,
                          const char *format, va_list args);
   bool general_log_write(THD *thd, enum enum_server_command command,
@@ -1320,22 +1372,26 @@ enum enum_binlog_format {
 int query_error_code(THD *thd, bool not_killed);
 uint purge_log_get_error_code(int res);
 
-int vprint_msg_to_log(enum loglevel level, const char *format, va_list args);
-void sql_print_error(const char *format, ...);
-void sql_print_warning(const char *format, ...);
-void sql_print_information(const char *format, ...);
-void sql_print_information_v(const char *format, va_list ap);
-typedef void (*sql_print_message_func)(const char *format, ...);
+int vprint_msg_to_log(enum loglevel level, const char *format, va_list args)
+  ATTRIBUTE_FORMAT(printf, 2, 0);
+void sql_print_error(const char *format, ...) ATTRIBUTE_FORMAT(printf, 1, 2);
+void sql_print_warning(const char *format, ...) ATTRIBUTE_FORMAT(printf, 1, 2);
+void sql_print_information(const char *format, ...)
+  ATTRIBUTE_FORMAT(printf, 1, 2);
+void sql_print_information_v(const char *format, va_list ap)
+  ATTRIBUTE_FORMAT(printf, 1, 0);
+typedef void (*sql_print_message_func)(const char *format, ...)
+  ATTRIBUTE_FORMAT_FPTR(printf, 1, 2);
 extern sql_print_message_func sql_print_message_handlers[];
 
 int error_log_print(enum loglevel level, const char *format,
-                    va_list args);
+                    va_list args) ATTRIBUTE_FORMAT(printf, 2, 0);
 
 bool slow_log_print(THD *thd, const char *query, uint query_length,
                     ulonglong current_utime);
 
 bool general_log_print(THD *thd, enum enum_server_command command,
-                       const char *format,...);
+                       const char *format,...) ATTRIBUTE_FORMAT(printf, 3, 4);
 
 bool general_log_write(THD *thd, enum enum_server_command command,
                        const char *query, size_t query_length);
@@ -1349,6 +1405,7 @@ File open_binlog(IO_CACHE *log, const char *log_file_name,
 
 void make_default_log_name(char **out, const char* log_ext, bool once);
 void binlog_reset_cache(THD *thd);
+void binlog_clear_incident(THD *thd);
 bool write_annotated_row(THD *thd);
 int binlog_flush_pending_rows_event(THD *thd, bool stmt_end,
                                     bool is_transactional,
@@ -1362,7 +1419,7 @@ binlog_cache_data* binlog_get_cache_data(binlog_cache_mngr *cache_mngr,
                                          bool use_trans_cache);
 
 extern MYSQL_PLUGIN_IMPORT MYSQL_BIN_LOG mysql_bin_log;
-extern handlerton *binlog_hton;
+extern transaction_participant binlog_tp;
 extern LOGGER logger;
 
 extern const char *log_bin_index;
@@ -1429,7 +1486,7 @@ end:
 
 static inline TC_LOG *get_tc_log_implementation()
 {
-  if (total_ha_2pc <= 1)
+  if (total_ha_2pc <= 2) // online_alter_tp and MHNSW_Trx::tp
     return &tc_log_dummy;
   if (opt_bin_log)
     return &mysql_bin_log;
@@ -1448,8 +1505,8 @@ const char *
 get_gtid_list_event(IO_CACHE *cache, Gtid_list_log_event **out_gtid_list);
 
 int binlog_commit(THD *thd, bool all, bool is_ro_1pc= false);
-int binlog_commit_by_xid(handlerton *hton, XID *xid);
-int binlog_rollback_by_xid(handlerton *hton, XID *xid);
+int binlog_commit_by_xid(XID *xid);
+int binlog_rollback_by_xid(XID *xid);
 bool write_bin_log_start_alter(THD *thd, bool& partial_alter,
                                uint64 start_alter_id, bool log_if_exists);
 #endif /* LOG_H */
