@@ -24,15 +24,13 @@
   classes to use when handling where clause
 */
 
-#ifdef USE_PRAGMA_INTERFACE
-#pragma interface			/* gcc class implementation */
-#endif
-
 #include "procedure.h"
 #include "sql_array.h"                        /* Array */
 #include "records.h"                          /* READ_RECORD */
 #include "opt_range.h"                /* SQL_SELECT, QUICK_SELECT_I */
 #include "filesort.h"
+#include "sql_delete.h"
+#include "sql_update.h"
 
 #include "cset_narrowing.h"
 
@@ -62,7 +60,7 @@ typedef struct keyuse_t {
     !NULL - This KEYUSE was created from an equality that was wrapped into
             an Item_func_trig_cond. This means the equality (and validity of 
             this KEYUSE element) can be turned on and off. The on/off state 
-            is indicted by the pointed value:
+            is indicated by the pointed value:
               *cond_guard == TRUE <=> equality condition is on
               *cond_guard == FALSE <=> equality condition is off
 
@@ -380,8 +378,9 @@ typedef struct st_join_table {
   */
   table_map	dependent;
   /*
-    key_dependent is dependent but add those tables that are used to compare
-    with a key field in a simple expression. See add_key_field().
+    Normally `key_dependent` is the same as `dependent` but may also include
+    tables that are used to compare with a key field in a simple expression
+    (see add_key_field()).
     It is only used to prune searches in best_extension_by_limited_search()
   */
   table_map     key_dependent;
@@ -558,6 +557,13 @@ typedef struct st_join_table {
   /** HAVING condition for checking prior saving a record into tmp table*/
   Item *having;
 
+  /**
+    Ordering to be produced when doing full index scan.
+    Important for vector indexes, set by test_if_skip_sort_order() when it
+    decides to use full index to produce rows in order.
+  */
+  ORDER *full_index_scan_order;
+
   /** TRUE <=> remove duplicates on this table. */
   bool distinct;
 
@@ -800,6 +806,12 @@ public:
 
   virtual void mark_used() = 0;
 
+  /*
+     Returns TRUE if the strategy is disabled by either optimizer switch
+     setting or an optimizer hint
+  */
+  virtual bool is_disabled() const { return false; }
+
   virtual ~Semi_join_strategy_picker() = default;
 };
 
@@ -820,12 +832,15 @@ class Duplicate_weedout_picker : public Semi_join_strategy_picker
   table_map dupsweedout_tables;
   
   bool is_used;
+
+  bool disabled; // See comment for Semi_join_strategy_picker::is_disabled()
 public:
   void set_empty() override
   {
     dupsweedout_tables= 0;
     first_dupsweedout_table= MAX_TABLES;
     is_used= FALSE;
+    disabled= FALSE;
   }
   void set_from_prev(POSITION *prev) override;
   
@@ -836,10 +851,13 @@ public:
                  double *record_count,
                  double *read_time,
                  table_map *handled_fanout,
-                 sj_strategy_enum *stratey,
+                 sj_strategy_enum *strategy,
                  POSITION *loose_scan_pos) override;
 
   void mark_used() override { is_used= TRUE; }
+
+  bool is_disabled() const override { return disabled; }
+
   friend void fix_semijoin_strategies_for_picked_join_order(JOIN *join);
 };
 
@@ -1021,7 +1039,7 @@ public:
     pushed down selection condition is applied) per each row combination of
     previous tables.
 
-    In best_access_path() it is set to the minum number of accepted rows
+    In best_access_path() it is set to the minimum number of accepted rows
     for any possible access method or filter:
 
     records_out takes into account table->cond_selectivity, the WHERE clause
@@ -1129,7 +1147,7 @@ public:
     *very* imprecise guesses made in best_access_path().
   */
   bool use_join_buffer;
-  /* True if we can use join_buffer togethere with firstmatch */
+  /* True if we can use join_buffer together with firstmatch */
   bool firstmatch_with_join_buf;
   POSITION();
 };
@@ -1161,7 +1179,7 @@ class Pushdown_query;
 
   @details
     The result records are obtained on the put_record() call.
-    The aggrgation process is determined by the write_func, it could be:
+    The aggregation process is determined by the write_func, it could be:
       end_write          Simply store all records in tmp table.
       end_write_group    Perform grouping using join->group_fields,
                          records are expected to be sorted.
@@ -1290,7 +1308,7 @@ public:
   */
   TABLE    **table;
   /**
-    The table which has an index that allows to produce the requried ordering.
+    The table which has an index that allows to produce the required ordering.
     A special value of 0x1 means that the ordering will be produced by
     passing 1st non-const table to filesort(). NULL means no such table exists.
   */
@@ -1370,7 +1388,7 @@ public:
 
   /*
     Tables one is allowed to use in choose_plan(). Either all or
-    set to a mapt of the tables in the materialized semi-join nest
+    set to a map of the tables in the materialized semi-join nest
   */
   table_map allowed_tables;
 
@@ -1405,10 +1423,10 @@ public:
   table_map eq_ref_tables;
 
   table_map allowed_top_level_tables;
-  ha_rows  send_records,found_records,join_examined_rows, accepted_rows;
+  ha_rows  send_records,found_records, accepted_rows;
 
   /*
-    LIMIT for the JOIN operation. When not using aggregation or DISITNCT, this 
+    LIMIT for the JOIN operation. When not using aggregation or DISTINCT, this
     is the same as select's LIMIT clause specifies.
     Note that this doesn't take sql_calc_found_rows into account.
   */
@@ -1442,6 +1460,7 @@ public:
 
   Pushdown_query *pushdown_query;
   JOIN_TAB *original_join_tab;
+  uint	   original_table_count;
   uint	   sort_space;
 
 /******* Join optimization state members start *******/
@@ -1501,7 +1520,7 @@ public:
   double   best_read;
   /*
     Estimated result rows (fanout) of the join operation. If this is a subquery
-    that is reexecuted multiple times, this value includes the estiamted # of
+    that is reexecuted multiple times, this value includes the estimated # of
     reexecutions. This value is equal to the multiplication of all
     join->positions[i].records_read of a JOIN.
   */
@@ -1606,7 +1625,7 @@ public:
 
   bool need_tmp; 
   bool hidden_group_fields;
-  /* TRUE if there was full cleunap of the JOIN */
+  /* TRUE if there was full cleanup of the JOIN */
   bool cleaned;
   DYNAMIC_ARRAY keyuse;
   Item::cond_result cond_value, having_value;
@@ -1625,7 +1644,7 @@ public:
 
     Then, ORDER/GROUP BY and Window Function code add columns that need to
     be saved to be available in the post-group-by context. These extra columns
-    are added to the front, because this->all_fields points to the suffix of
+    are added to the front, because this->fields_list points to the suffix of
     this list.
   */
   List<Item> all_fields;
@@ -1648,13 +1667,13 @@ public:
   COND *conds;                            // ---"---
   Item *conds_history;                    // store WHERE for explain
   COND *outer_ref_cond;       ///<part of conds containing only outer references
-  COND *pseudo_bits_cond;     // part of conds containing special bita
+  COND *pseudo_bits_cond;     // part of conds containing special bits
   TABLE_LIST *tables_list;           ///<hold 'tables' parameter of mysql_select
   List<TABLE_LIST> *join_list;       ///< list of joined tables in reverse order
   COND_EQUAL *cond_equal;
   COND_EQUAL *having_equal;
   /*
-    Constant codition computed during optimization, but evaluated during
+    Constant condition computed during optimization, but evaluated during
     join execution. Typically expensive conditions that should not be
     evaluated at optimization time.
   */
@@ -1736,6 +1755,13 @@ public:
   */
   bool is_orig_degenerated;
 
+  /*
+    DELETE and UPDATE may have an imitation JOIN, which is not NULL,
+    but has NULL join_tab. In such cases we may want to access
+    sql_cmd_dml::scanned_rows to choose optimization strategies.
+  */
+  Sql_cmd_dml *sql_cmd_dml;
+
   JOIN(THD *thd_arg, List<Item> &fields_arg, ulonglong select_options_arg,
        select_result *result_arg)
     :fields_list(fields_arg)
@@ -1784,7 +1810,7 @@ public:
   bool make_sum_func_list(List<Item> &all_fields, List<Item> &send_fields,
 			  bool before_group_by);
 
-  /// Initialzes a slice, see comments for ref_ptrs above.
+  /// Initializes a slice, see comments for ref_ptrs above.
   Ref_ptr_array ref_ptr_array_slice(size_t slice_num)
   {
     size_t slice_sz= select_lex->ref_pointer_array.size() / 5U;
@@ -1872,11 +1898,7 @@ public:
                            ulonglong curr_space,
                            ulonglong needed_space);
   void set_allowed_join_cache_types();
-  bool is_allowed_hash_join_access()
-  { 
-    return MY_TEST(allowed_join_cache_types & JOIN_CACHE_HASHED_BIT) &&
-           max_allowed_join_cache_level > JOIN_CACHE_HASHED_BIT;
-  }
+  bool is_allowed_hash_join_access(const TABLE *table);
   /*
     Check if we need to create a temporary table.
     This has to be done if all tables are not already read (const tables)
@@ -1934,6 +1956,10 @@ public:
   bool optimize_upper_rownum_func();
   void calc_allowed_top_level_tables(SELECT_LEX *lex);
   table_map get_allowed_nj_tables(uint idx);
+  bool propagate_dependencies(JOIN_TAB *stat);
+  void update_key_dependencies();
+  table_map *export_table_dependencies() const;
+  void restore_table_dependencies(table_map *orig_dep_array);
 
 private:
   /**
@@ -2018,7 +2044,7 @@ int opt_sum_query(THD* thd,
                   List<TABLE_LIST> &tables, List<Item> &all_fields, COND *conds);
 
 /* from sql_delete.cc, used by opt_range.cc */
-extern "C" int refpos_order_cmp(void* arg, const void *a,const void *b);
+extern "C" int refpos_order_cmp(void *arg, const void *a,const void *b);
 
 /** class to copying an field/item to a key struct */
 
@@ -2414,6 +2440,25 @@ public:
     for (uint i= 0; i < s->fields; i++)
       field[i]->set_null();
   }
+
+  /*
+    Run the event handler for all fields in the table
+    in the range [start, end-1].
+  */
+  void expr_event_handler(THD *thd, expr_event_t event, uint start, uint end)
+  {
+    DBUG_ASSERT(start <= end);
+    DBUG_ASSERT(end <= s->fields);
+    for (uint i= start; i < end; i++)
+      field[i]->expr_event_handler(thd, event);
+  }
+
+  // Run the event handler for all fields in the table
+  void expr_event_handler(THD *thd, expr_event_t event)
+  {
+    expr_event_handler(thd, event, 0, s->fields);
+  }
+
   /**
     Set all fields from a compatible item list.
     The number of fields in "this" must be equal to the number
@@ -2485,7 +2530,7 @@ create_virtual_tmp_table(THD *thd, List<Spvar_definition> &field_list)
     is enabled, we now enable "simulate_out_of_memory". This effectively
     makes table->init() fail on OOM inside multi_alloc_root().
     This is done to test that ~Virtual_tmp_table() called from the "delete"
-    below correcly handles OOM.
+    below correctly handles OOM.
   */
   DBUG_EXECUTE_IF("simulate_create_virtual_tmp_table_out_of_memory",
                   DBUG_SET("+d,simulate_out_of_memory"););
@@ -2709,5 +2754,12 @@ void propagate_new_equalities(THD *thd, Item *cond,
                               COND_EQUAL *inherited,
                               bool *is_simplifiable_cond);
 
+#define PREV_BITS(type, N_BITS) ((type)my_set_bits(N_BITS))
+
+double estimate_post_group_cardinality(JOIN *join, double join_output_card);
+
 bool dbug_user_var_equals_str(THD *thd, const char *name, const char *value);
+
+#include "opt_vcol_substitution.h"
+
 #endif /* SQL_SELECT_INCLUDED */
